package org.lukeallen.spiderrobot;


import gnu.io.CommPortIdentifier;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.UnsupportedCommOperationException;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Timer;
import java.util.TimerTask;

public class robotControl {

    RobotInterface window;
    
    public int[][] liveservotimes = new int[6][3];
    public int[][] servopluslimits = new int[6][3];
    public int[][] servominuslimits = new int[6][3];
    //the below arrays will be used to calculate pulse length given angle, in the form y=mx+b
    //a.k.a. pulse=gain*angle + zero
    public double[][] servozero = new double[6][3];//pulse length that produces angle=0
    public double[][] servogain = new double[6][3];//change in pulse length per change in degree
    //other control and calibration variables
    int calPulse1 = 0; //length of pulse that yielded calAngle1, according to the user.
    int calAngle1 = -999;
    int calPulse2 = 0;
    int calAngle2 = -999;
    public static final int mainloopperiod = 50;//time delay between runs of the main loop, in ms.
    //code used for the test motion sequence
    double elapsed;//seconds passed since last iteration through main operating loop
    boolean movementSequenceOn = false;
    int sequenceStep; //where the leg is in the movement sequence
    double sequenceTimer;//timer to next step in sequence

    //serial stuff:
    static Enumeration	      portList;
    static CommPortIdentifier portId;
    static SerialPort	      serialPort;
    static OutputStream       outputStream;
    static boolean	      outputBufferEmptyFlag = false;

    //the constructor takes my user input window as an argument
    //It starts running the mainloop() method every 50ms until the application closes.
    //mainloop() polls the window for inputs and responds to them.
    robotControl(RobotInterface inputwindow) {
        window = inputwindow;

        openSerial();//set up the serial port; right now it's hard-coded to COM1. Need to add interface.

        Timer timer = new Timer();
        timer.schedule(new TimerTask() {
            public void run() {
                mainloop();
            }
        }, 250,//delay before first run; doesn't really matter
                mainloopperiod);//repeat period in ms

        boolean calibrated = loadCalibration();//read in cal. data from file
        if(calibrated){
        	for(int i=0; i<liveservotimes.length;i++){
                for(int j=0; j<liveservotimes[i].length; j++){
                    liveservotimes[i][j] = (servominuslimits[i][j]+servopluslimits[i][j])/2;//set initial pulse length to middle of calibrated band.
                }
            }
        }else{//not calibrated yet, so we'll have to make assumptions.
	        for(int i=0; i<liveservotimes.length;i++){
	            for(int j=0; j<liveservotimes[0].length; j++){
	                liveservotimes[i][j] = 128;//set default pulse length to middle of possible band, as this is unlikely to hurt a servo.
	            }
	        }
        }
    }



    //mainloop. This is run repeatedly by the timer set up in main().
    //It handles overall control of the robot.
    public void mainloop(){
    	elapsed = (double)mainloopperiod/1000;//assumes we're executing at our requested rate. Replace this with a timer if that becomes a bad assumption. 
        
        //-----------handle calibration if enabled-------------
        if(window.calOn() == true){
        	livecalibrate();
        }
        else{//if cal is off, clear any cal-related commands or else they'll be acted on all at once later.
            window.saveCal = false;
            window.setPlusLim = false;
            window.setMinusLim = false;
            window.calAngle = -999;
        }

        //--------------handle leg position testing command-------
        if(window.testMove == true){
            window.testMove = false;//reset button variable
            movementSequenceOn = false;//cancel any sequences in progress
            moveLegTo(window.getLeg(), window.getTestPosition());
            
        }

        //--------------handle single step command------------
        if(window.takeStep == true){
            window.takeStep = false;//reset button variable
            movementSequenceOn = true; //start the step sequence
            sequenceStep = 0;
            sequenceTimer = 0;
        }

        if(movementSequenceOn){
        	boolean finished = continueSequence();
        	if(finished)
        		movementSequenceOn = false;
        }
        window.setText2("zero: " + servozero[window.getLeg()][window.getServo()] + " gain: "
                + servogain[window.getLeg()][window.getServo()]);
        window.setText3(" -limit: " + servominuslimits[window.getLeg()][window.getServo()]
                + "+limit: " + servopluslimits[window.getLeg()][window.getServo()]);

        transmitToMicrocontroller();
    }
    
    //This executes a movement sequence. For now, it makes Leg 0 move its foot in a circle as it's repeatedly called during the MOL. Returns true when complete.
    boolean continueSequence(){
    	int sequenceTotalSteps = 20;//steps to divide the circle into 
    	double sequenceTotalTime = .7;//seconds to complete sequence
    	//foot positions in cm:
    	double centerX = -32;
    	double centerY = 6;
    	double centerZ = -12;
    	double radius = 7; 
    	sequenceTimer -= elapsed;
    	if(sequenceTimer <= 0){//we need to advance to next position
    		sequenceTimer = sequenceTotalTime/sequenceTotalSteps;//timer to next repositioning
    		//if(sequenceStep == 0)
    		//	sequenceTimer += 1;//add a 1 second delay to get to first position
    		//next position in the test circle in the constant Z plane:
    		double radians = (double)sequenceStep*(2*Math.PI)/sequenceTotalSteps;
    		Position pos = new Position();
    		pos.x = centerX+Math.sin(radians)*radius;
    		pos.y = centerY+Math.cos(radians)*radius;
    		pos.z = centerZ;
    		
    		moveLegTo(0, pos);
    		sequenceStep++;//this will be the next position
    	}
    	if(sequenceStep >= sequenceTotalSteps){
    		sequenceStep=0;
    		return true;//sequence is finished
    	}
    	else
    		return false;//not finished yet
    }
    /*constant X circle demo sequence:
      	double centerX = -32;
    	double centerY = 6;
    	double centerZ = -4;
    	double radius = 10; 
    	pos.x = centerX;
		pos.y = centerY+Math.cos(radians)*radius;
		pos.z = centerZ+Math.sin(radians)*radius;
    */
    
    
    
    //reposition the servo in response to user commands, and take user's report of the resulting joint angle.
    private void livecalibrate(){
    	liveservotimes[window.getLeg()][window.getServo()] = window.getSlider();//send the pulse width to the servo
        //display commanded angles
        double dispangle = (window.getSlider()-servozero[window.getLeg()][window.getServo()])/servogain[window.getLeg()][window.getServo()];
        window.setText4(Math.round(dispangle) + " deg");

        if(window.saveCal == true){
            saveCalibration();
            window.saveCal = false;
            window.setText("Calibration data saved to file servocal.txt");
        }

        if(window.setPlusLim == true){
            servopluslimits[window.getLeg()][window.getServo()] = window.getSlider();
            window.setPlusLim = false;
        }
        if(window.setMinusLim == true){
            servominuslimits[window.getLeg()][window.getServo()] = window.getSlider();
            window.setMinusLim = false;
        }



        if(window.calAngle != -999){//means calibration angle has just been set
            if(calAngle1 == -999){//it hasn't been set
                calAngle1 = window.calAngle;
                calPulse1 = window.getSlider();
                window.calAngle = -999;//read the calibration angle and reset the input
                window.setText("first point saved, set new angle");
            }
            else{//angle 1 has been set, but angle 2 hasn't
                calAngle2 = window.calAngle;
                calPulse2 = window.getSlider();
                window.calAngle = -999;
                double gain = (double)(calPulse2 - calPulse1)/(calAngle2 - calAngle1);//find slope
                double zero = (double)calPulse1 - (calAngle1 * gain);//solve for b in pulse=m*angle + b
                servozero[window.getLeg()][window.getServo()] = zero;
                servogain[window.getLeg()][window.getServo()] = gain;
                calAngle1 = -999;
                calAngle2 = -999;//reset to be ready for a new calculation
                window.setText("Zero and gain calculated.");
            }
        }
    }

    
    private void moveLegTo(int leg, Position command){
        JointAngles angles  = findJointAngles(leg, command);
        servoCommand theServoCommand = servoCommandFromAngles(leg, angles);
        if(theServoCommand.possible == false){
            window.setText("Impossible leg movement command received");
            return;
        }
        for(int i=0; i<3; i++){
            liveservotimes[leg][i] = theServoCommand.servopulse[i];
        }
        window.setText("Leg moved");
        //will probably need to add handling of intermediate positions at some point
    }

    static double legshoulderx[] = new double[6];//shoulder positions
    static double legshouldery[] = new double[6];
    static double legshoulderz[] = new double[6];//this actually needs to be the height of the joint at the end of segment 1
    static double seg1length[] = new double[6];//length of the part before the "up" joint
    static double seg2length[] = new double[6];
    static double seg3length[] = new double[6];

    private JointAngles findJointAngles(int leg, Position command){

        //temporarily hacking this in. It will be read from a file eventually.
        legshoulderx[0] = 0;
        legshouldery[0] = 0;
        legshoulderz[0] = 0;
        seg1length[0] = 8;
        seg2length[0] = 13.8;
        seg3length[0] = 22.6;






        //first change command values into a reference frame where the shoulder of this leg is 0,0,0
        double x = command.x - legshoulderx[leg];
        double y = command.y - legshouldery[leg];
        double z = command.z - legshoulderz[leg];
        double angle1 = Math.atan2(x,y);//this is arctan(x/y). I'll unconventionally define 0 radians as the +y direction.
        System.out.println(angle1*180/Math.PI);
        //note there's a discontinuity at the -y direction which would mess up the angle-to-pulse-length conversion
        //below. But my legs never have to move through that direction, so I won't bother accounting for it.

        double R = Math.sqrt(x*x + y*y);//distance from shoulder to commanded point, neglecting z
        R = R - seg1length[leg];//really though, we care about the distance that has to be covered by the remaining 2 segments
        //now find the two vertical joint angles using Equations 1.7 and 1.8 of "Robot Modeling and Control" book
        //which derived this using the law of cosines. (Replace 'x' with R and 'y' with z in the equations.)
        double a1 = seg2length[leg];
        double a2 = seg3length[leg];
        double D = (R*R+z*z-a1*a1-a2*a2)/(2*a1*a2);
        double angle3 = Math.atan2(-1*Math.sqrt(1-D*D), D);//This is the "elbow up" solution, which I'll use for the spider.
                                                            //the "elbow down" solution is given by removing the "-1"
        double angle2 = Math.atan2(z, R) - Math.atan2( a2*Math.sin(angle3), a1 + a2*Math.cos(angle3));

        System.out.println("angle1: " + Math.round(angle1));
        System.out.println(angle1*180/Math.PI);
        System.out.println("angle2: " + Math.round(angle2));
        System.out.println(angle2*180/Math.PI);
        System.out.println("angle3: " + Math.round(angle3));
        System.out.println(angle3*180/Math.PI);
        return new JointAngles(angle1, angle2, angle3);
    }
    
    private servoCommand servoCommandFromAngles(int leg, JointAngles angles){
    	servoCommand theServoCommand = new servoCommand();
    	theServoCommand.servopulse[0] = (int)Math.round(angles.a1*180/Math.PI*servogain[leg][0] + servozero[leg][0]);
        theServoCommand.servopulse[1] = (int)Math.round(angles.a2*180/Math.PI*servogain[leg][1] + servozero[leg][1]);
        theServoCommand.servopulse[2] = (int)Math.round(angles.a3*180/Math.PI*servogain[leg][2] + servozero[leg][2]);
        theServoCommand.possible = true;
        for(int i = 0; i<3; i++){
            if(theServoCommand.servopulse[i]>servopluslimits[leg][i] ||
                    theServoCommand.servopulse[i]<servominuslimits[leg][i]){
                theServoCommand.possible = false;
                System.out.println("Requires movement beyond limit: Pulselength " + i +" was:");
                System.out.println(theServoCommand.servopulse[i]);
            }
        }
        return theServoCommand;
    }

    private void saveCalibration() {
        try {
            PrintWriter outputStream = new PrintWriter(new FileWriter("servocal.txt"));
            outputStream.println("//Servo config file. Lines preceded by / will be ignored. File is wiped with each calibration.");
            for(int i=0;i<servopluslimits.length;i++){
                outputStream.println("//Leg " + i);
                for(int j=0; j<servopluslimits[0].length;j++){
                    outputStream.println("//Servo " + j + " + limit:");
                    outputStream.println(servopluslimits[i][j]);
                    outputStream.println("//- limit:");
                    outputStream.println(servominuslimits[i][j]);
                    outputStream.println("//zero:");
                    outputStream.println(servozero[i][j]);
                    outputStream.println("//gain:");
                    outputStream.println(servogain[i][j]);
                }
            }
            outputStream.close();
        }
        catch (IOException e) {
            window.setText("File problem; could not save.");
        }
    }


    private boolean loadCalibration() {
        try {
            BufferedReader inputStream = new BufferedReader(new FileReader("servocal.txt"));
            String line = null;

            for(int i=0;i<servopluslimits.length;i++){
                for(int j=0; j<servopluslimits[0].length;j++){
                    while((line = inputStream.readLine()).charAt(0)=='/')//ignore the comment lines
                    {}
                    servopluslimits[i][j] = Integer.parseInt(line);
                    while((line = inputStream.readLine()).charAt(0)=='/')
                    {}
                    servominuslimits[i][j] = Integer.parseInt(line);
                    while((line = inputStream.readLine()).charAt(0)=='/')
                    {}
                    servozero[i][j] = Double.parseDouble(line);
                    while((line = inputStream.readLine()).charAt(0)=='/')
                    {}
                    servogain[i][j] = Double.parseDouble(line);
                    }
                }
            inputStream.close();
            return true;
            }
        catch (IOException e) {
            window.setText("Could not load previous servo calibration data.");
            return false;
        }
    }

    private void openSerial(){
        boolean portFound = false;
	String  defaultPort = "COM1";

	portList = CommPortIdentifier.getPortIdentifiers();

	while (portList.hasMoreElements()) {
	    portId = (CommPortIdentifier) portList.nextElement();

	    if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {

		if (portId.getName().equals(defaultPort)) {
		    System.out.println("Found port " + defaultPort);

		    portFound = true;

		    try {
			serialPort =
			    (SerialPort) portId.open("SimpleWrite", 2000);
		    } catch (PortInUseException e) {
			System.out.println("Port in use.");

			continue;
		    }

		    try {
			outputStream = serialPort.getOutputStream();
		    } catch (IOException e) {}

		    try {
			serialPort.setSerialPortParams(9600,
						       SerialPort.DATABITS_8,
						       SerialPort.STOPBITS_1,
						       SerialPort.PARITY_NONE);
		    } catch (UnsupportedCommOperationException e) {}

		    try {
		    	serialPort.notifyOnOutputEmpty(true);
		    } catch (Exception e) {
			System.out.println("Error setting event notification");
			System.out.println(e.toString());
			System.exit(-1);
		    }
		}
	    }
	}

	if (!portFound) {
	    System.out.println("port " + defaultPort + " not found.");
	}
    }

    public void closeSerial(){
        serialPort.close();
    }
    
    
    //transmitToMicrocontroller() is a custom method that talks to my microcontroller code to give it the desired servo pulse times.
    //for efficiency I express them as 0-255, which the microcontroller will scale to 0.5-2.5 ms repeating every 20ms
    private void transmitToMicrocontroller(){
    	//----forward servo pulse times to the microcontroller, to be applied to the servos.
    	//System.out.println("Writing \""+messageString+"\" to "
    	//		+serialPort.getName());
        byte[] transmission = new byte[liveservotimes.length * liveservotimes[0].length];
        for(int i=0; i<liveservotimes.length;i++){
            for(int j=0; j<liveservotimes[0].length; j++){
                transmission[i*liveservotimes[0].length + j] = (byte)liveservotimes[i][j];//I'll need to redo this if I want more than 0-255
            }
        }
		try {
			outputStream.write("go".getBytes());//This signifies the start of the message string.
	                                                    //(The microcontroller may get turned on in the middle, so it needs to get synced up.)
	                outputStream.write(transmission);
		} catch (IOException e){
		}
    }


}

class Position{//This stores a desired robot foot position. (My code assumes the position is expressed in the robot's coordinate frame)
    public double x;
    public double y;
    public double z;

    Position(){

    }
    Position(double inputx, double inputy, double inputz){
        x = inputx;
        y = inputy;
        z = inputz;
    }
}
class JointAngles{//This stores joint angles of one leg
	public double a1;
    public double a2;
    public double a3;

    JointAngles(){

    }
    JointAngles(double a1in, double a2in, double a3in){
        a1 = a1in;
        a2 = a2in;
        a3 = a3in;
    }
}
  //This contains the pulse widths for one robot leg worth of servos, as well as
  //a boolean stating whether the move can be done, given leg lengths and
  //joint angle limits.
class servoCommand{
	 public int servopulse[];
	 public boolean possible;
	 servoCommand(){
		 servopulse = new int[3];
	 }
}